#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netdb.h>

/* ORDER::::
    domain names that are accept/reject
    ip addresses that are accept/reject
    matched ip/sender accepted by file
    matched ip/sender accepted (implicit)
    Matched absolute accept/reject addresses. Needed because
	prodigy is not very consistent (IE Meridian.com)
	[for rejects, lets blacklist the IP too]
	Note we still use the access.db list in postfix...
    grey listed
*/

#define IP_ACCEPT 1
#define IP_GREY 0
#define IP_REJECT -1

typedef struct ipTerminal {
	char mode;     /* Each individual address + global */
	char *user;
	char *ipName;
	unsigned createTime;
	} ipTerminal;
typedef struct ipLink {
	union {
	    struct ipLink *nextLink;
	    struct ipTerminal *terminal;
	    } ipx[257];
	} ipLink;

ipLink *ipRoot;

typedef struct addrInfo {
	char *domainPattern;
	short mode;	/* IP_ACCEPT/IP_REJECT */
	struct addrInfo *next;
	} addrInfo;
addrInfo *addrRoot=NULL;

/* Checked after domain ip checks/network pattern checks. Verfies a mail from
   a sender say foo@yahoo.com comes from a machine on yahoo's network 
   Note this check is done *before* a grey list, but after existing up/down ip/addr checks */
typedef struct senderCheckDomain {
	char *senderDomainPattern;
	char *networkDomainPattern;
	struct senderCheckDomain *next;
	} senderCheckDomain;
senderCheckDomain *senderRoot = NULL;

typedef struct senderFroms {
	char *fromPattern;
	short mode;
	struct senderFroms *next;
	} senderFroms;
senderFroms *fromRoot = NULL;

/* Msg passing. From postfix, looks like IP addr, from domain, from usr */
/* Msg back is just len of 1 byte with value "A", accept, "G" grey,
   "I" for IP reject, "D" for domain reject or "M" for mismatch imposter
   "S" for sender address reject */
/* Msg from postfix is always 2048 bytes for convenience */

time_t greyTime=60;
int portNum=2525;	/* Default port to listen on */
int debug=0;
int refresh = 0;
int dirty = 0;
struct pollfd pfds[256];

busage(){
	printf("usage: addressDB <staticFile> <dynamicFile> [port|-interactive] [-d]\n");
	printf("HUP to reread info files immediately\n");
	exit(44);
}

freeDynamic()
{
int i1, i2, i3, i4;
ipLink *l1, *l2, *l3;
ipTerminal *t4;

    for(i1=0; i1<256; i1++)
	{
	if( !(l1=ipRoot->ipx[i1].nextLink) )
	    continue;
	for(i2=0; i2<256; i2++)
	    {
	    if( (l2=l1->ipx[256].nextLink) )
		{
		free(l2);
		l1->ipx[256].nextLink = NULL;
		printf("Free %d.X\n", i1);
		}
	    if( !(l2=l1->ipx[i2].nextLink) )
		continue;
	    for(i3=0; i3<256; i3++)
		{
		if( (l3=l2->ipx[256].nextLink) )
		    {
		    free(l3);
		    l2->ipx[256].nextLink = NULL;
		    printf("Free %d.%d.X\n", i1, i2);
		    }
		if( !(l3=l2->ipx[i3].nextLink) )
		    continue;
		for(i4=0; i4<256; i4++)
		    {
		    if( (t4=l3->ipx[256].terminal) )
			{
			free(t4);
			l3->ipx[256].nextLink = NULL;
			printf("Free %d.%d.%d.X\n", i1, i2, i3);
			}
		    if( !(t4=l3->ipx[i4].terminal) )
			continue;
		    free(t4);
		    l3->ipx[i4].terminal = NULL;
		    }
		}
	    }
	}
}

freeAll()
{
senderCheckDomain *s1, *s2;
addrInfo *a1, *a2;
senderFroms *f1, *f2;

    freeDynamic();
    for(f1=fromRoot; f1; f1=f2)
	{
	f2 = f1->next;
	free(f1->fromPattern);
	free(f1);
	}

    for(a1=addrRoot; a1; a1=a2)
	{
	a2 = a1->next;
	free(a1->domainPattern);
	free(a1);
	}

    for(s1=senderRoot; s1; s1=s2)
	{
	s2 = s1->next;
	free(s1->senderDomainPattern);
	free(s1->networkDomainPattern);
	free(s1);
	}
    addrRoot = NULL;
    senderRoot = NULL;
}

sameSenderDomain(char *s1, char *s2)
{
int l1, l2;

    l1 = strlen(s1);
    l2 = strlen(s2);

    for(l1--, l2--; l1>=0 && l2>=0; l1--, l2--)
	if( s1[l1] != s2[l2] )
	    return(0);
	else if( s1[l1] == '@' )
	    return(1);
    return(1);
}

/* Does the from address match the pattern ? */
fromstrcmp(char *from, char *fromPattern)
{
int pIx, sIx;

    sIx = strlen(from);
    pIx = strlen(fromPattern);

    /* Pattern *must* be shorter */
    if( pIx > sIx )
	return(0);

    for(sIx--, pIx--; pIx>=0; sIx--, pIx--)
	if( from[sIx] != fromPattern[pIx] )
	    {
	    if( from[sIx] != '@' )
		return(0);
	    }
    if( from[sIx] == '@' || from[sIx] == '.' )
	return(1);
    return(0);
}

/* Not really a pattern, but a backwards strcmp sort of */
netstrcmp(char *ipName, char *ipPattern)
{
int pIx, sIx;

    sIx = strlen(ipName);
    pIx = strlen(ipPattern);
    /* Pattern *must* be shorter */
    if( pIx > sIx )
	return(0);

    for(sIx--, pIx--; pIx>=0; sIx--, pIx--)
	if( ipName[sIx] != ipPattern[pIx] )
	    return(0);

    return(1);
}

breakIP(char *ip, short *addr)
{
char *s1, *s2;
int i;

    for(s1=ip, i=0; i<4; i++)
	{
	if( *s1 )
	    {
	    addr[i] = atoi(s1);
	    for(; *s1; s1++)
		if( *s1 == '.' )
		    { s1++; break; }
	    }
	else
	    addr[i] = 256;
	}
}

/* Query only probes... */
int queryIP(short *addr)
{
int i;
ipLink *l, *ll;
ipTerminal *t;

    l = ipRoot;
    for(i=0; i<4; i++)
	{
	if( (t=l->ipx[256].terminal) )
	    return(t->mode);
	if( i == 3 )   /* Must be terminal */
	    {
	    if( !(t=l->ipx[addr[i]].terminal) )
		return(0);
	    return(t->mode);
	    }
	else
	    {
	    if( !(ll=l->ipx[addr[i]].nextLink) )
		ll=l->ipx[addr[i]].nextLink = (ipLink *) calloc(sizeof(ipLink), 1);
	    l = ll;
	    }
	}
}
int defineIP(short *addr, int mode, unsigned curTime, char *usr, char *ipName)
{
int i;
ipLink *l, *ll;
ipTerminal *t;

    l = ipRoot;
    for(i=0; i<4; i++)
	{
	if( mode == 0 ) /* Query */
	    {   /* And a global was set */
	    if( (t=l->ipx[256].terminal) )
		{
		return(t->mode);
		}
	    }

	if( addr[i] == 256 )    /* global */
	    {
	    if( !(t=l->ipx[256].terminal) )
		{
		t = l->ipx[256].terminal = (ipTerminal *) calloc(sizeof(ipTerminal), 1);
		t->mode = mode;
		return;
		}
	    }
	else if( i == 3 )   /* Must be terminal */
	    {
	    if( !(t=l->ipx[addr[i]].terminal) )
		{
		t = l->ipx[addr[i]].terminal = (ipTerminal *) calloc(sizeof(ipTerminal), 1);
		t->createTime = curTime;
		t->user = strdup(usr);
		t->ipName = strdup(ipName);
		}
	    if( mode == 0 ) /* A query */
		{
		if( t->mode == 0 )
		    {   /* Same user ? */
		    if( !sameSenderDomain(usr, t->user) )
			{   /* No, block now */
			if( debug )
			    printAction(addr, usr, ipName, "GREY->BLACK", -1);
			dirty = 1;
			t->mode = IP_REJECT;
			}
		    else if(t->createTime+greyTime < curTime )
			{   /* Are we past ok time ? */
			if( debug )
			    printAction(addr, usr, ipName, "GREY->WHITE", 1);
			dirty = 1;
			t->mode = IP_ACCEPT;
			}
		    }
		return(t->mode);
		}
	    if( debug )
		printAction(addr, usr, ipName, "Define IP mode as", mode);
	    dirty = 1;
	    t->mode = mode; /* Mode changes */
	    return(mode);
	    }
	else
	    {
	    if( !(ll=l->ipx[addr[i]].nextLink) )
		ll=l->ipx[addr[i]].nextLink = (ipLink *) calloc(sizeof(ipLink), 1);
	    l = ll;
	    }
	}
}

readStatic(char *fileName)
{
char line[8192], s1[8192], s2[8192], s3[8192];
short addr[4];
int cnt, mode;
FILE *fp;
senderCheckDomain *sNode;
addrInfo *addrNode;
senderFroms *fromNode;

    if( (fp=fopen(fileName, "r")) == NULL)
	{
	printf("File open fail static map %s\n", fileName);
	exit(44);
	}
    while(fgets(line, 8190, fp))
	{
	if(line[0] == '#' || isspace(line[0]) )
		continue;
	cnt = sscanf(line, "%s %s %s", s1, s2, s3);
	if( cnt == 2 )	/* ip# or domain straight up reject/accept */
	    {
	    if( strcmp(s2, "REJECT") == 0)
		mode = IP_REJECT;
	    else if( strcmp(s2, "TRUST") == 0)
		mode = IP_ACCEPT;
	    else
		{
		if( strcmp(s2, "ACCEPT_SENDER") == 0)
		    {
		    fromNode = (senderFroms *) calloc(sizeof(senderFroms), 1);
		    fromNode->fromPattern = strdup(s1);
		    fromNode->mode = IP_ACCEPT;
		    fromNode->next = fromRoot;
		    fromRoot = fromNode;
		    }
		else if( strcmp(s2, "REJECT_SENDER") == 0)
		    {
		    fromNode = (senderFroms *) calloc(sizeof(senderFroms), 1);
		    fromNode->fromPattern = strdup(s1);
		    fromNode->mode = IP_REJECT;
		    fromNode->next = fromRoot;
		    fromRoot = fromNode;
		    }
		continue;       /* skip MATCH_NET now */
		}

	    if( s1[0] == '@' )
		{
		breakIP( &(s1[1]), addr);
		defineIP(addr, mode, 0, "pre", "pre");
		}
	    else
		{
		addrNode = (addrInfo *) calloc(sizeof(addrInfo), 1);
		addrNode->mode = mode;
		addrNode->domainPattern = strdup(s1);
		addrNode->next = addrRoot;
		addrRoot = addrNode;
		}
	    }
	else if( cnt == 3 )
	    {
	    if( strcmp(s2, "MATCH_SENDER") == 0)
		{
		sNode = (senderCheckDomain *) calloc(sizeof(senderCheckDomain), 1);
		sNode->senderDomainPattern = strdup(s1);
		sNode->networkDomainPattern = strdup(s3);
		sNode->next = senderRoot;
		senderRoot = sNode;
		}
	    }
	}
    fclose(fp);
}

	
readDynamic(char *fileName)
{
char line[8192];
char addrS[64], modeS[10], timeS[20], userS[1024], domainS[1024];
char *s;
short addr[4];
int mode;
unsigned long createTime;
FILE *fp;

    fp = fopen(fileName, "r");
    if( !fp )
	return;

    while(fgets(line, 8190, fp))
	{
	if( sscanf(line, "%s %s %s %s %s", addrS, modeS, timeS, userS, domainS) != 5 )
	    continue;
	breakIP(addrS, addr);
	mode = atoi(modeS);
	createTime = strtoul(timeS, &s, 10);
	defineIP(addr, mode, createTime, userS, domainS);
	}
    fclose(fp);
}

/* Dynamics are all individual addresses */
writeDynamic(char *fileName)
{
int i1, i2, i3, i4;
ipLink *l1, *l2, *l3;
ipTerminal *t4;
FILE *fp;

    if( fileName == NULL )
	fp = stdout;
    else
	fp = fopen(fileName, "w");

    for(i1=0; i1<256; i1++)
	{
	if( !(l1=ipRoot->ipx[i1].nextLink) )
	    continue;
	for(i2=0; i2<256; i2++)
	    {
	    if( !(l2=l1->ipx[i2].nextLink) )
		continue;
	    for(i3=0; i3<256; i3++)
		{
		if( !(l3=l2->ipx[i3].nextLink) )
		    continue;
		for(i4=0; i4<256; i4++)
		    {
		    if( !(t4=l3->ipx[i4].terminal) )
			continue;
		    fprintf(fp, "%d.%d.%d.%d %d %u %s %s\n",
			i1, i2, i3, i4,
			t4->mode, t4->createTime, t4->user, t4->ipName);
		    }
		}
	    }
	}
    if( fileName )
	fclose(fp);
}

printAction(short addr[4], char *usr, char *ipName, char *reason, int mode)
{

    printf("%s %d IP %d.%d.%d.%d %s %s\n",
	reason, mode, addr[0], addr[1], addr[2], addr[3],
	    ipName, usr);
}

isSenderMatch(char *usr, char *ipName)
{
int l1, l2;

    l1 = strlen(usr);
    l2 = strlen(ipName);

    for(l1--, l2--; l1>=0 && l2>=0; l1--, l2--)
	{
	if( usr[l1] == '@' )
	    return(1);
	if( usr[l1] != ipName[l2] )
	    return(0);
	}
    if( l1 > 0 && usr[l1] == '@' )
	return(1);
    return(0);
}

processAddr(short addr[4], unsigned curTime, char *usr, char *ipName)
{
addrInfo *a1;
senderCheckDomain *x1;
senderFroms *f1;
int mode;
char *s;

    /* Lower case them */
    for(s=usr; *s; s++)
	if( isupper(*s) )
	    *s = tolower(*s);

    for(s=ipName; *s; s++)
	if( isupper(*s) )
	    *s = tolower(*s);

    /* Do we trust/reject the domain name ? */
    for(a1=addrRoot; a1; a1=a1->next)
	{
	if(netstrcmp(ipName, a1->domainPattern))
	    {
	    if( debug )
		printAction(addr, usr, ipName, "DNAME", a1->mode);
	    if( a1->mode == IP_REJECT )
		return('D');
	    return('A');
	    }
	}
    /* Do we trust/reject because of the IP address ? */
    mode = queryIP(addr);
    if( mode == IP_REJECT )
	{
	if( debug )
	    printAction(addr, usr, ipName, "DIP", -1);
	return('I');
	}
    if( mode == IP_ACCEPT )
	{
	if( debug )
	    printAction(addr, usr, ipName, "DIP", 1);
	return('A');
	}

    /* Do we trust it because it matched ? */
    for(x1=senderRoot; x1; x1=x1->next)
	{
	if( fromstrcmp(usr, x1->senderDomainPattern) )
	    {   /* Matched, must be from the specified network */
	    if(netstrcmp(ipName, x1->networkDomainPattern))
		mode = 'A';
	    else
		{
		mode = 'M';
		/* And further, blacklist it! */
		defineIP(addr, IP_REJECT, curTime, usr, ipName);
		}
	    if( debug )
		printAction(addr, usr, ipName, "TABMATCH", mode);
	    return(mode);
	    }
	}
    if( isSenderMatch(usr, ipName) )
	{
	if( debug )
	    printAction(addr, usr, ipName, "IMPMATCH", 1);
	return('A');
	}
    /* Is the sender's from address in the list */
    for(f1=fromRoot; f1; f1=f1->next)
	{
	if( fromstrcmp(usr, f1->fromPattern) )
	    {
	    if( f1->mode == IP_ACCEPT )
		{
		if( debug )
		    printAction(addr, usr, ipName, "OK_SENDER", 1);
		return('A');
		}
	    /* And further, blacklist it! */
	    if( debug )
		printAction(addr, usr, ipName, "REJECT_SENDER", -1);
	    defineIP(addr, IP_REJECT, curTime, usr, ipName);
	    return('S');
	    }
	}
    /* Do we need to greylist ? or do we know ? */
    mode = defineIP(addr, IP_GREY, curTime, usr, ipName);
    if( mode == 0 )
	return('G');
    if( mode == IP_ACCEPT )
	return('A');
    return('I');
}

doInteractive(char *staticFile, char *dynamicFile)
{
char line[256];
char s1[50], s2[128], s3[128];
short addr[4];
time_t curTime;
int mode;

    while(1)
    {
    printf("10.10.10.10 mail.yahoo.com user@yahoo.com\n  ");
    fflush(stdout);
    gets(line);
    if( strlen(line) < 8 )
	{
	if( strcasecmp(line, "reset") == 0)
	    {
	    printf("Write dynamic\n");
	    writeDynamic(dynamicFile);
	    printf("Free old data\n");
	    freeAll();
	    printf("Reload static\n");
	    readStatic(staticFile);
	    printf("Reload dynamic\n");
	    readDynamic(dynamicFile);
	    continue;
	    }
	writeDynamic(NULL);
	continue;
	}
    sscanf(line, "%s %s %s", s1, s2, s3);
    time(&curTime);
    breakIP(s1, addr);
    mode = processAddr(addr, curTime, s3, s2);
    printf("RESULT: %c\n", mode);
    }
}

static void sigrefresh(int signum)
{
    refresh = 1;
}

handleMsg(int i)
{
char msg[2048], reply[16];
char *ip, *domain, *usr;
int fetched, cnt, z;
short addr[4];
time_t t;
unsigned curTime;

    for(fetched=cnt=0; fetched<2048 && cnt < 100; cnt++)
	{
	z = read(pfds[i].fd, &msg[fetched], 2048-fetched);
	if( z == -1 )
	    {
	    if( errno == EAGAIN || errno == EINTR )
		{
		usleep(10000);
		continue;
		}
	    printf("Closing socket index %d; read %d bytes so far\n", i, fetched);
	    close(pfds[i].fd);
	    pfds[i].fd = -1;
	    return;
	    }
	fetched += z;
	if( fetched < 2048 )
	    usleep(10000);
	else
	    break;
	}

    /* Break up msg */
    ip = msg;
    for(z=0; z<2048; z++)
	if( msg[z] == '\0' )
	    break;

    z++;
    if( z >= 2048 )
	goto bad;
    domain = &msg[z];
    for(; z<2048; z++)
	if( msg[z] == '\0' )
	    break;
    z++;
    if( z >= 2048 )
	goto bad;
    usr = &msg[z];

    time(&t);
    curTime = t;
    breakIP(ip, addr);
    reply[0]=processAddr(addr, curTime, usr, domain);
    reply[1] = '\0';

    write(pfds[i].fd, reply, 16);
    return;

bad:
    printf("Bad msg on descriptor %d\n", i);
    close(pfds[i].fd);
    pfds[i].fd = -1;
}

main(int argc, char *argv[])
{
char *staticFile=NULL, *dynamicFile=NULL;
int fd, i, j, cnt, rc;
int numActive=1;

    for(cnt=0, i=1; i<argc; i++)
	{
	if( strcasecmp(argv[i], "-d") == 0)
		debug = 1;
	else if( strcasecmp(argv[i], "-interactive") == 0)
		portNum = 0;
	else if( cnt == 0 )
		{
		cnt++;
		staticFile = argv[i];
		}
	else if( cnt == 1 )
		{
		cnt++;
		dynamicFile = argv[i];
		}
	else if( cnt == 2 )
		{
		cnt++;
		portNum = atoi(argv[i]);
		}
	else if( cnt == 3 )
		{
		greyTime = atoi(argv[i]);
		cnt++;
		}
	else
		cnt++;
	}
    if( cnt < 2 || cnt > 4 )
	busage();

    ipRoot = (ipLink *) calloc(sizeof(ipLink), 1);

    readStatic(staticFile);
    readDynamic(dynamicFile);
    if( portNum == 0 )
	doInteractive(staticFile, dynamicFile);
    openPort(portNum, &fd);
    pfds[0].fd = fd;
    pfds[0].events = POLLIN | POLLOUT;

    signal(SIGPIPE, SIG_IGN);
    signal(SIGINT, sigrefresh);

    /* Wait for connections */
again:
    i = poll(pfds, numActive, 30000);
    if( i < 0 )
	{
	if( refresh )   /* Good time to do it */
	    {
	    freeAll();
	    readStatic(staticFile);
	    readDynamic(dynamicFile);
	    refresh = 0;
	    }
	else if( dirty )
	    {
	    writeDynamic(dynamicFile);
	    printf("Flushing\n");
	    dirty = 0;
	    }
	goto again;
	}
    if( i == 0 )
	goto again;

    if( pfds[0].revents & POLLIN )
	{
	struct   sockaddr from;
	socklen_t len;

	printf("Accept started\n");
	len = sizeof(struct   sockaddr);
	rc = accept(pfds[0].fd, &from, &len);
	printf("Accept returns %d\n", rc);
	if( rc >= 0 )
	    {
	    for(j=1; j<numActive; j++)
		if( pfds[j].fd < 0 )
		    break;
	    pfds[j].fd = rc;
	    pfds[j].events = POLLIN | POLLHUP;
	    pfds[j].revents = 0;
	    if( j == numActive )
		numActive++;
	    }
	goto again;
	}
    for(j=1; j<numActive; j++)
	{
	if( pfds[i].revents & POLLHUP )
	    {
	    printf("pollhup?\n");
	    exit(44);
	    }
	if( pfds[i].revents & POLLERR )
	    {
	    printf("pollerr?\n");
	    exit(44);
	    }
	if( pfds[i].revents & POLLNVAL)
	    {
	    printf("pollnva?\n");
	    exit(44);
	    }
	if( pfds[i].revents & POLLIN)
	    handleMsg(i);
	}
    for( ; numActive>1; numActive--)
	if( pfds[numActive-1].fd != -1 )
	    break;

    goto again;
}

openPort(int portNum, int *fdp)
{
int fd;
char me[MAXHOSTNAMELEN];
struct hostent  *h;
struct sockaddr_in in;
unsigned int x;
char *fix, ss;

    gethostname(me, MAXHOSTNAMELEN);
    fd = socket(PF_INET, SOCK_STREAM, 0);

    h = gethostbyname(me);
    x = INADDR_ANY;
    memcpy( &(in.sin_addr), (void *) &x, sizeof(struct in_addr));
    in.sin_family = AF_INET;
    in.sin_port = portNum;

#if 0
    fix = (char *) &in.sin_port;
    ss = fix[0];
    fix[0] = fix[1];
    fix[1] = ss;
#endif

    if( bind(fd, (const struct sockaddr *)&in, sizeof(in)) != 0 )
	{
	printf("Failed to bind socket\n");
	exit(44);
	}
    if( listen(fd, 5) != 0)
	{
	printf("Failed to listen\n");
	exit(44);
	}

    *fdp = fd;
}
